#include "bitstring.h"

/* Access levels */

#define ALEV_CONF	0
#define ALEV_ITEM	1
#define ALEV_RESP	2
#define NLEV		3

/* These need to be indexes into the table defined in access.c */

#define ACCESS_MAYREAD		 0
#define ACCESS_MAYPOST		 1
#define ACCESS_MAYNAME		 2
#define ACCESS_MAYLOCK		 3
#define ACCESS_MAYLINKIN	 4

#define ACCESS_MAYLINKOUT	 5
#define ACCESS_MAYRESP		 6
#define ACCESS_MAYKILL		 7
#define ACCESS_MAYRETIRE	 8
#define ACCESS_MAYRETITLE	 9
#define ACCESS_MAYFREEZE	10

#define ACCESS_MAYHIDE		11
#define ACCESS_MAYERASE		12
#define ACCESS_MAYEDIT		13

#define NACCESS 		14


/* ACCESS DESCRIPTORS are images of 'access' files loaded into memory.  We
 * usually have one for the current conference, and sometimes have some for
 * other conferences (e.g., when we are linking an item from another
 * conference).
 *
 * The "Access" structure is a linked list of rules, listed in the same order
 * they appeared in in the access file.  Rules that are known at load time
 * to always be false for the current user are dropped from the list.  Each
 * rule contains bit strings identifying which access types (from the above
 * list) it sets.
 *
 * When we read in an 'access' file, we always already know who the user is.
 * So many of the conditions can be evaluated at read time, like "fwlist" and
 * "registered".  If a rule has only such conditions, then we just have to
 * save we know if it is true or false.  Always false rules are dropped from
 * the linked list.  Always true rules have a proc Array of size 0.  Other
 * rules depend on characteristics of the current item or response.  For these
 * we save a backtalk expression that, when evaluated, tells us if it is true
 * or not.
 *
 */

struct access_rule {
    Array *proc;	/* If not NULL, a procedure to evaluate the rule */
    bit_decl(set, NACCESS)	/* Access bits to set if true */
    bit_decl(val, NACCESS)	/* Values to set access bits to if true */
    struct access_rule *next;
    };

typedef struct access_rule *Access;

/* Function declarations for access.c */

int find_access_type(char *name);
int rule_fires (Access rule);
int maydo(Access acc, int what);
Access new_rule(void);
void free_rule(Access rule);
int access_word(char *word, Access rule, int *procmem);
Access read_access(char *dir);
